<?php
namespace PrestaShop\Module\PolarisPrestaConnector\pol;

use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\ProduitDetailTaille;
use PrestaShop\Module\PolarisPrestaConnector\BridgeAdapter;
use PrestaShop\Module\PolarisPrestaConnector\Models\Article;
use PrestaShop\Module\PolarisPrestaConnector\Models\AttributeRef;
use PrestaShop\Module\PolarisPrestaConnector\SyncConfiguration;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\CatalogApi;
use PrestaShop\Module\PolarisPrestaConnector\Models\Product;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\ConfigApi;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\InfoConnect;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Model\Pager;
//use PrestaShopBundle\Translation\Api\StockApi;
use PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Api\StocksApi;
use PrestaShop\Module\PolarisPrestaConnector\Models\Stock;

/**
 * Pont vers Polaris
 * 
 * Memo : utilisation de l'API Polaris possible ici 
 * $config = Configuration::getDefaultConfiguration()->setApiKey('X-API-Key', 'XX XXXX XXXX XXXX XXXX XXXX XXXX XXXX');
 * $config->getDefaultConfiguration()->setHost('https://XXX:14243/api/WebConnect');
 * $api = new CatalogApi(null, $config);
 */
class Bridge extends BridgeAdapter
{
    /**
     * Nombre de produits à récupérer en une fois
     */
    private const NB_PRODUCTS_AT_A_TIME = 100;

    /**
     * Nombre de stocks à récupérer en une fois
     */
    private const NB_STOCKS_AT_A_TIME = 10000;

    /**
     * URL de base de l'API
     */
    private string $baseUrl = "";

    /**
     * Informations de connexion
     */
    private ?InfoConnect $structure = null;

    /**
     * Cache de nom de grilles de tailles
     */
    private array $sizeGrids = [];

    /**
     * Api Polaris de produit
     */
    private ?CatalogApi $catalogApi = null; 

    /**
     * Api Polaris de stock
     */
    private ?StocksApi $stocksApi = null;

    /**
     * Dernière position de produits synchronisés
     */
    private string|null $lastProductPosition = null;

    /**
     * Dernière position de stocks synchronisés
     */
    private string|null $lastStockPosition = null;

    /**
     * Constructeur
     */
    public function __construct(SyncConfiguration $config)
    {
        parent::__construct($config);        
    }

    /**
     * Obtient la configuration de connexion pour l'API
     */
    private function getApiConfig() : \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Configuration
    { 
        // On charge la configuration
        $loginCfg = $this->config->get(SyncConfiguration::CFG_LOGIN);

        if (!isset($loginCfg) || !isset($loginCfg['address']) || !isset($loginCfg['key']))
        {
            if (!empty($loginCfg))
            {
                $loginCfg = null;
                $this->config->set(SyncConfiguration::CFG_LOGIN, $loginCfg);  // reset !
            }

            throw new \Exception("Le module n'est pas associé au backoffice Polaris !");
        }

        $config = \PrestaShop\Module\PolarisPrestaConnector\pol\api\lib\Configuration::getDefaultConfiguration();

        $config->setHost($loginCfg['address'] .  '/api/WebConnect');
        $config->setApiKey('X-API-Key', $loginCfg['key']);

        $this->baseUrl = $loginCfg['address'] . '/';

        return $config;
    }

    /**
     * Obtient les informations de l'API
     */
    private function getStructure() : InfoConnect
    {        
        if ($this->structure === null)
        {
            // Les infos n'ont pas encore été tirées, on s'en occupe !
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                return null;

            $infoApi = new ConfigApi(null, $apiCfg);
            $this->structure = $infoApi->configGetStruture();
            if (!$this->structure)
                throw new \Exception('Impossible de récupérer les informations de connexion');

            // BUG API Polaris : grilles de tailles
            $this->sizeGrids = [];
            foreach ($this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS) as $grid => $size)
                $this->sizeGrids[$grid] = $size;
        }

        return $this->structure;
    }

    /**
     * Renvoi l'identifiant du pont correspondant au backoffice géré
     */
    public function getId() : string
    {
        return 'Polaris';
    }

    /**
     * Renvoi les spécificités du pont sous forme de tableau associatif clé/valeur de type string
     * 
     * @return array<string,string>
     */
    public function getSpecifics() : array
    {
        //  La connexion avec votre backoffice LCV a été perdue. Merci de recommencer l'association.
        return [            
            "ADDRESS_LABEL" => "Quel est votre numéro de licence (il est de la forme pl-pX-XXXX-XXX)",
        ];
    }

    /**
     * Importe la configuration du backoffice via les API
     */
    public function pullEnvironment()
    {
        // Pour la suite, on va devoir tout charger, ça sera quand même bc plus rapide !
        if (!$this->config->hasBeenPreloaded())
            $this->config->preload();

        // Charger la structure
        $structure = $this->getStructure();

        // Fixation du magasin web
        if ($this->config->get(SyncConfiguration::CFG_WEB_STORE_CODE) !== $structure->getMagasin()->getCode())
            $this->config->set(SyncConfiguration::CFG_WEB_STORE_CODE, $structure->getMagasin()->getCode());

        // Nom des catégories dynamiques
        $dyncat = $this->config->get(SyncConfiguration::CFG_DYN_CAT); 
        $dynattr = $this->config->get(SyncConfiguration::CFG_DYN_ATTR); 

        if (!$this->catalogApi)
            $this->catalogApi = new CatalogApi(null, $this->getApiConfig());

        // Les catégories dynamiques
        foreach (['categorie1', 'categorie2', 'categorie3', 'categorie4'] as $cat)
        {
            $funcName = 'get' . ucfirst($cat);
            $catname = $structure->$funcName()->getNom();
            if (!isset($dyncat[$cat]) || $dyncat[$cat] != $catname)
                $this->config->setMap(SyncConfiguration::CFG_DYN_CAT, $cat, $catname);
        } 

        // Les attributs dynamiques
        foreach (['niveau1', 'niveau2', 'niveau3', 'niveau4'] as $attr)
        {
            $funcName = 'get' . ucfirst($attr);
            $attrname = $structure->$funcName()->getNom();
            if (!isset($dynattr[$cat]) || $dynattr[$cat] != $attrname)
                $this->config->setMap(SyncConfiguration::CFG_DYN_ATTR, $attr, $attrname);
        }

        // Tirage des valeurs des catégories dynamiques
        $empty = [ "*" => 0 ];
        $emptyName = [ "*" => "Tous les autres" ];
        foreach ([  'Categories1' => 'categorie1', 
                    'Categories2' => 'categorie2',
                    'Categories3' => 'categorie3',
                    'Categories4' => 'categorie4',
                    'Collections' => 'collection',
                    'Saisons' => 'saison'
                    ] as $func => $cat)
        {
            $funcName = 'catalogGet' . ucfirst($func);
            $mapName = SyncConfiguration::CFG_MAP_DYN_CAT . $cat;
            if ($this->config->get($mapName) === null)  /* auto création */
                $this->config->set($mapName, $empty);
            if ($this->config->get($mapName.'_name') === null)  /* auto création */
                $this->config->set($mapName.'_name', $emptyName);

            // On tire les catégorie
            $pager = null;
            do
            {                
                $pageCategories = $this->catalogApi->$funcName(null, null, json_encode($pager));
                $pager = $pageCategories->getPagerNext();                

                foreach ($pageCategories->getItems() as $category)
                {
                    $code = $category->getNo();
                    $name = $category->getNom();
                    if ($this->config->getMap($mapName, $code) === null)
                        $this->config->setMap($mapName, $code, 0, true);

                    if ($this->config->getMap($mapName."_name", $category->getNo()) != $name)
                        $this->config->setMap($mapName."_name", $category->getNo(), $name);
                }
            }
            while ($pager != null);
        }

        // Tirage RFS
        $mapName = SyncConfiguration::CFG_MAP_DYN_CAT . 'rfs';
        if ($this->config->get($mapName) === null)  /* auto création */
            $this->config->set($mapName, $empty);
        if ($this->config->get($mapName.'_name') === null)  /* auto création */
            $this->config->set($mapName.'_name', $emptyName);

        // On tire les RFS
        $pager = null;
        do
        {                
            $pageCategories = $this->catalogApi->catalogGetClassifications(null, null, null, null, json_encode($pager));
            $pager = $pageCategories->getPagerNext();                

            foreach ($pageCategories->getItems() as $category)
            {
                $code = $category->getCode();
                $name = $category->getCode() . ' - ' . $category->getNom();
                if ($this->config->getMap($mapName, $code) === null)
                    $this->config->setMap($mapName, $code, 0, true);

                if ($this->config->getMap($mapName."_name", $code) != $name)
                    $this->config->setMap($mapName."_name", $code, $name);
            }
        }
        while ($pager != null);

        // Charger les grilles de tailles
        // BUG POLARIS : on ne peux pas tirer les grilles de tailles, alors on va les chercher en live (voir code dans nextProducts)
        
        // On tire les marques
        if ($this->config->get(SyncConfiguration::CFG_MAP_MANUFACTURER) === null)
            $this->config->set(SyncConfiguration::CFG_MAP_MANUFACTURER, $empty);

        $pager = null;
        do
        {
            $pageMarques = $this->catalogApi->catalogGetMarques(null, null, json_encode($pager));
            $pager = $pageMarques->getPagerNext();

            foreach ($pageMarques->getItems() as $marque)
            {
                if ($this->config->getMap(SyncConfiguration::CFG_MAP_MANUFACTURER, $marque->getNo()) === null)
                    $this->config->setMap(SyncConfiguration::CFG_MAP_MANUFACTURER, $marque->getNo(), 0, true);

                if ($this->config->getMap(SyncConfiguration::CFG_MAP_MANUFACTURER_NAME, $marque->getNo()) != $marque->getNom())
                    $this->config->setMap(SyncConfiguration::CFG_MAP_MANUFACTURER_NAME, $marque->getNo(), $marque->getNom());
            }
        }
        while ($pager != null);
    }

    /**
     * Révoque les configurations non supportées par le pont
     * 
     * @param array $cfg Configuration à révoquer
     */
    public function revokeUnsupportedCfg(array & $cfg)
    {
        // A priori, nous supportons toutes les options proposées
        // sauf l'écotaxe ?
        unset($cfg[SyncConfiguration::CFG_PRODUCT_SYNC_ECOTAX]);            // L'écotaxe n'est pas supportée
    }

    /**
     * Configuration par défaut de ce pont
     * 
     */
    public function alterDefaultCfg(array & $cfg)
    {
        $cfg[SyncConfiguration::CFG_DYN_CAT] = [
            'rfs' => 'RFS',
            'categorie1' => 'Catégorie 1',
            'categorie2' => 'Catégorie 2',
            'categorie3' => 'Catégorie 3',
            'categorie4' => 'Catégorie 4',            
            'collection' => 'Collection',
            'saison' => 'Saison', 
        ];

        $cfg[SyncConfiguration::CFG_DYN_ATTR] = [
            'niveau1' => 'Niveau 1',
            'niveau2' => 'Niveau 2',
            'niveau3' => 'Niveau 3',
            'niveau4' => 'Niveau 4',
        ];

        $cfg[SyncConfiguration::CFG_PR_ATTRIBUTES_MAPPING] = [
            SyncConfiguration::PRODUCT_ATTR_VARIANT => [
                'niveau1' => 0,
                'niveau2' => 0, // souvent la couleur
                'niveau3' => 0,
                'niveau4' => 0,
            ],
            SyncConfiguration::PRODUCT_ATTR_PRODUCT => [
                'niveau2' => 0, // souvent la couleur
            ],
            SyncConfiguration::PRODUCT_ATTR_FEATURE => [
                'collection' => 0, 
                'saison' => 0
            ],
        ];
    }

    /**
     * Teste la connexion au backoffice avec les paramètres fournis
     */
    public function bind($args) : string|null
    {
        // Interressons nous à l'adresse qui est pour Polaris le numéro de licence
        // avec ce numéro, on va en déduire l'adresse de connexion
        
        if (!isset($args["address"]) || empty(trim($args["address"])))
            return "1 : le numéro de licence n'est pas défini !";

        $licenceId = strtolower(trim($args["address"]));

        // Le numéro de licence Polaris doit être de la forme "pl-pX-XXXX-XXX" où X est un chiffre ou "pl-vega-xxxx" où x est un chiffre, quelque soit le nombre de chiffres pour les deux formes
        if (!preg_match('/^pl-(p\d-)?\d+-\d+$/', $licenceId) && !preg_match('/^pl-vega-\d+$/', $licenceId))
            return "2 : le numéro de licence n'est pas valide !";        

        // On peut retrouver l'URL de connexion à partir du numéro de licence en se connectant au gestionnaire de licence polaris
        // avec l'URL suivante : https://extranet.vega-info.fr/manager/api/Public/polaris/XX/Url où XX est le numéro de licence (url escapé)
        // Cet endpoint nous renvoi un JSON avec l'adresse de connexion dans la propriété "url"
        $urlManager = 'https://extranet.vega-info.fr/manager/api/Public/polaris/' . urlencode($licenceId) . '/Url';
        $response = file_get_contents($urlManager);
        if ($response === false)
            return "3 : impossible de contacter le gestionnaire de licence Polaris. Merci de réessayer dans quelques instants. Si l'erreur se reproduit, merci de contacter le support technique de VEGA Informatique.";
        $url = json_decode($response)->url;

        // On vérifie maintenant que le pincode est correct. Il est rangé sous password
        if (!isset($args["pincode"]) || empty(trim($args["pincode"])))
            return "30 : renseignez maintenant le pincode de jumelage. Il est disponible dans votre backoffice Polaris."; 

        // On vérifie maintenant que le pincode est correct.
        // On va se connecter à l'API Polaris en POST avec les informations fournies pour transformer notre pincode en clé d'accès
        // Si on obtient un 401, c'est que le pincode est incorrect. Il nous faut donc le code de réponse HTTP du serveur
        // Pour cela, on va plutôt utiliser curl en GET   
        $ch = curl_init($url . '/api/Core/App/Binding?pinCode=' . urlencode($args["pincode"]));
        curl_setopt($ch, CURLOPT_HTTPGET, true);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true); // Permet de suivre les redirections
        curl_setopt($ch, CURLINFO_HEADER_OUT, true);
        $response = curl_exec($ch);
        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        curl_close($ch);

        if ($httpCode == 0)
            return "31 : impossible de contacter le backoffice Polaris. Vérifiez que vos ports sont ouverts sur votre box et que votre TLR est allumé !";

        if ($httpCode == 401)
            return "32 : le pincode de jumelage est incorrect. Merci de vérifier le code fourni.";

        if ($httpCode != 200)
            return "33 : une erreur inconnue (".$httpCode.") est survenue. Merci de contacter le support technique de VEGA Informatique.";

        // Si on arrive ici, on a toutes les informations nécessaires
        $response = json_decode($response, true);

        // sauver la config
        $loginCfg = [
            'address' => $url,
            'key' => $response["Guid"] . '/' . $response["Secret"],
        ];
        $this->config->set(SyncConfiguration::CFG_LOGIN, $loginCfg);

        return null;
    }
    
    /**
     * Obtient l'URL d'une photo par sa clé
     */
    public function getPhotoURL(string $photo) : string
    {
        return $this->baseUrl . sprintf("api/Files/Media/Photo?noPhoto=%s&photoSize=original", $photo);
    }

    /**
     * Obtient le code magasin internet
     */
    public function getMagasinInternet() : string
    {
        $codeMagasin = $this->config->get(SyncConfiguration::CFG_WEB_STORE_CODE);
        if (!$codeMagasin)
        {
            $this->pullEnvironment();   
            $codeMagasin = $this->config->get(SyncConfiguration::CFG_WEB_STORE_CODE);
            if (!$codeMagasin)
                throw new \Exception('Erreur de configuration : le magasin web n\'est pas défini');
        }

        return $codeMagasin;
    }

    /**
     * Obtient les produits à synchroniser sous forme de tableau
     * 
     * @return array<Product>
     */
    public function & nextProducts() : array
    {
        if ($this->catalogApi === null)
        {
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                throw new \Exception('Impossible de se connecter à l\'API Polaris');

            $this->catalogApi = new CatalogApi(null, $apiCfg);
        }
        
        $products = [];
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastProductPosition)
        {            
            $this->setSyncPosition('products', $this->lastProductPosition);
            $this->lastProductPosition = null;
        }

        // Données de base de l'API BUG Grilles de taille
        $this->sizeGrids = $this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS);
        
        // Construction du pager
        $pager = new Pager();        
        $pager->setFrom($this->getSyncPosition('products') ?? '');
        $pager->setNb(self::NB_PRODUCTS_AT_A_TIME);
        $pager->setSort('+DerniereModification, +No');        
        
        // endless pager ?
        $pager = json_encode($pager); // bug OpenAPI

        $codeMagasin = $this->getMagasinInternet();

        $pageProducts = $this->catalogApi->catalogGetProduits(null, null, $pager, null, 'true', 'true', 'true', 'true' );
        if ($pageProducts === null || $pageProducts->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les produits');

        foreach ($pageProducts->getItems() as $source)
        {
            // Construction du produit
            $product = new Product($source->getNoModele(), $source->getCode());

            // Attributs du produit
            $product->productAttributes = [
                "categorie1" => new AttributeRef($source->getCategorie1()->getNo(), $source->getCategorie1()->getNom()),
                "categorie2" => new AttributeRef($source->getCategorie2()->getNo(), $source->getCategorie2()->getNom()),
                "categorie3" => new AttributeRef($source->getCategorie3()->getNo(), $source->getCategorie3()->getNom()),
                "categorie4" => new AttributeRef($source->getCategorie4()->getNo(), $source->getCategorie4()->getNom()),

                "niveau1" => new AttributeRef($source->getNiveau1()->getNo(), $source->getNiveau1()->getNom()),
                "niveau2" => new AttributeRef($source->getNiveau2()->getNo(), $source->getNiveau2()->getNom()),
                "niveau3" => new AttributeRef($source->getNiveau3()->getNo(), $source->getNiveau3()->getNom()),
                "niveau4" => new AttributeRef($source->getNiveau4()->getNo(), $source->getNiveau4()->getNom()),

                "collection" => new AttributeRef($source->getCollection()->getNo(), $source->getCollection()->getNom()),
                "saison" => new AttributeRef($source->getSaison()->getNo(), $source->getSaison()->getNom()),
                "rfs" => new AttributeRef($source->getClassification()->getCode(), $source->getClassification()->getNom()),
            ];

            // Grille de tailles
            $product->grilleTailles = $source->getGrilleTaille();

            // BUG API Polaris : on ne peux pas tirer les grilles de tailles, alors on va les chercher en live
            if (!array_key_exists($product->grilleTailles, $this->sizeGrids))
            {
                $this->sizeGrids = $this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS);
                $this->sizeGrids[$product->grilleTailles] = 0;
                $this->config->set(SyncConfiguration::CFG_MAP_SIZEGRIDS, $this->sizeGrids);

                $sizeGridsName = $this->config->get(SyncConfiguration::CFG_MAP_SIZEGRIDS_NAME);
                $sizeGridsName[$product->grilleTailles] = $product->grilleTailles;
                $this->config->set(SyncConfiguration::CFG_MAP_SIZEGRIDS_NAME, $sizeGridsName);
            }

            // Taxe
            $product->taxrate = $source->getTauxTva();

            // Description
            $product->description = $source->getDescription();

            // Les dimensions / poids
            $product->weight = $source->getPoids();
            $volume = $source->getVolume();
            $product->depth = $volume[0] ?? null;
            $product->width = $volume[1] ?? null;
            $product->height = $volume[2] ?? null;

            $product->ecotax = $source->getEcoParticipationHt();

            // Les déclinaisons (tailles)            
            foreach ($source->getTailles() as $taille)
            {
                $refs = $taille->getRefs();
                if (empty($refs)) continue;

                $article = new Article($refs[0], $taille->getTaille());

                if (!empty($taille->getEans()))
                    $article->ean13 = $taille->getEans()[0];

                // Prix TTC
                foreach ($taille->getMagasins() as $store)
                    if ($store->getCodeMagasin() == $codeMagasin)
                        $article->priceTTC = $store->getPrixNormalTtc();

                // Zonage
                $article->warehouseLocation = $taille->getZonage();

                $product->articles[] = $article;
            }
            
            // Nom du produit
            $product->name = $source->getLibelle();

            // Marque
            $product->manufacturer = new AttributeRef($source->getMarque()->getNo(), $source->getMarque()->getNom());

            // Photos
            $product->photos = $source->getPhotos() ?? [];

            // Stock
            foreach ($source->getTailles() as $taille)
                $product->stocks[] = $this->buildStockFromTaille($taille);
            
            $products[] = $product;
        }        

        // Fini
        $this->lastProductPosition = $pageProducts->getEndlessPager()->getFrom();
        return $products;
    }

    /**
     * Construit un stock à partir d'une response ProduitDetailTaille
     * 
     * @param ProduitDetailTaille $taille Taille à convertir
     * @return Stock Stock construit
     */
    private function buildStockFromTaille(ProduitDetailTaille $taille) : Stock
    {
        $codeMagasin = $this->getMagasinInternet();

        $prixTTC = null;
        $qties = [];
        foreach ($taille->getMagasins() as $store)
        {
            $qties[$store->getCodeMagasin()] = $store->getDispo();

            // Si on est dans le magasin Internet, on met le prix à jour
            if ($store->getCodeMagasin() == $codeMagasin)
                $prixTTC = $store->getPrixNormalTtc();
        }
        
        $stock = new Stock($taille->getRefs(), $qties);

        // Ean ?
        $stock->eans = $taille->getEans();

        // Prix ?
        $stock->prixTTC = $prixTTC;
    
        return $stock;
    }

    /**
     * Obtient les stocks à synchroniser sous forme de tableau
     * 
     * @return array<Stock>
     */
    public function & nextStocks(): array
    {
        if ($this->stocksApi === null)
        {
            $apiCfg = $this->getApiConfig();
            if (!$apiCfg)
                throw new \Exception('Impossible de se connecter à l\'API Polaris');

            $this->stocksApi = new stocksApi(null, $apiCfg);
        }
        
        $stocks = [];
        // Si nous avons une ancienne position à sauver, c'est le moment
        if ($this->lastStockPosition)
        {            
            $this->setSyncPosition('stocks', $this->lastStockPosition);
            $this->lastStockPosition = null;
        }
        
        // Construction du pager
        $pager = new Pager();        
        $pager->setFrom($this->getSyncPosition('stocks') ?? '');
        $pager->setNb(self::NB_STOCKS_AT_A_TIME);
        $pager->setSort('+DerniereModification, +No');        
        
        // endless pager ?
        $pager = json_encode($pager); // bug OpenAPI        

        $pageStocks = $this->stocksApi->stocksGetStocks(null, null, $pager, null, 'false', 'false', 'false', 'false' );
        if ($pageStocks === null || $pageStocks->getItems() === null)
            throw new \Exception('Erreur API Polaris : impossible de récupérer les stocks');

        foreach ($pageStocks->getItems() as $source)
        {            
            foreach ($source->getTailles() as $taille)
                $stocks[] = $this->buildStockFromTaille($taille);
        }        

        // Fini
        $this->lastStockPosition = $pageStocks->getEndlessPager()->getFrom();
        return $stocks;
    }
}
